/*******************************************************************************
Copyright (C) Marvell International Ltd. and its affiliates

This software file (the "File") is owned and distributed by Marvell 
International Ltd. and/or its affiliates ("Marvell") under the following
alternative licensing terms.  Once you have made an election to distribute the
File under one of the following license alternatives, please (i) delete this
introductory statement regarding license alternatives, (ii) delete the two
license alternatives that you have not elected to use and (iii) preserve the
Marvell copyright notice above.

********************************************************************************
Marvell Commercial License Option

If you received this File from Marvell and you have entered into a commercial
license agreement (a "Commercial License") with Marvell, the File is licensed
to you under the terms of the applicable Commercial License.

********************************************************************************
Marvell GPL License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File in accordance with the terms and conditions of the General 
Public License Version 2, June 1991 (the "GPL License"), a copy of which is 
available along with the File in the license.txt file or by writing to the Free 
Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 or 
on the worldwide web at http://www.gnu.org/licenses/gpl.txt. 

THE FILE IS DISTRIBUTED AS-IS, WITHOUT WARRANTY OF ANY KIND, AND THE IMPLIED 
WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE ARE EXPRESSLY 
DISCLAIMED.  The GPL License provides additional details about this warranty 
disclaimer.
********************************************************************************
Marvell BSD License Option

If you received this File from Marvell, you may opt to use, redistribute and/or 
modify this File under the following licensing terms. 
Redistribution and use in source and binary forms, with or without modification, 
are permitted provided that the following conditions are met:

    *   Redistributions of source code must retain the above copyright notice,
	    this list of conditions and the following disclaimer. 

    *   Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution. 

    *   Neither the name of Marvell nor the names of its contributors may be 
        used to endorse or promote products derived from this software without 
        specific prior written permission. 
    
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR 
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON 
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/telephony.h>
#include <linux/phonedev.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/poll.h>
#include "proslic.h"

#ifdef MV_88F5181L
#include "mvTdm.h"
#define MP_IRQ	13
#else
#include "mvTdmFpga.h"
#define MP_IRQ	(32+3)
#endif

/****************
* DEBUG CONTROL *
****************/
/* debug main on/off switch (more in debug control below ) */
//#define MP_DEBUG
//#undef MP_DEBUG

#define MP_DBG_OFF     0x0000
#define MP_DBG_RX      0x0001
#define MP_DBG_TX      0x0002
#define MP_DBG_POLL    0x0004
#define MP_DBG_IOCTL   0x0008
#define MP_DBG_ISR     0x0010
#define MP_DBG_ALL     0xffff

#ifdef MP_DEBUG
# define MPDBG(FLG, X) if( (mp_dbg & (FLG)) == (FLG) ) printk X
#else
# define MPDBG(FLG, X)
#endif

unsigned int mp_dbg = MP_DBG_OFF; /* MP_DBG_ISR | MP_DBG_IOCTL; */ /*MP_DBG_ALL*/

/* Extra IOCTLs for debug (used by mv_voip_tool) */
#define MV_IOCTL
#ifdef MV_IOCTL
#define PHONE_MV_READ_REG	_IOWR ('q', 0xB0, unsigned short)
#define PHONE_MV_WRITE_REG	_IOW ('q', 0xB1, unsigned int)
#define PHONE_MV_SPI_TEST	_IO ('q', 0xB2)
#endif

/* Statistics counters */
#define MV_STAT
#ifdef MV_STAT
#define PHONE_MV_STAT_SHOW	_IO ('q', 0xB5)
#define MV_STAT_INC(counter)	((counter)++);
typedef struct {
	unsigned int read_req, read_buff_empty, read_blocking, read_copy_to_user;
	unsigned int write_req, write_buff_full, write_blocking, write_copy_from_user;
	unsigned int poll_req, poll_can_read, poll_can_write, poll_can_do_event;
	unsigned int int_total, int_slic, int_data, int_wake_poll, int_empty;
	unsigned int int_write_ok,int_write_sync_lost, int_write_buff_empty;
	unsigned int int_read_ok, int_read_sync_lost, int_read_buff_full;
	unsigned int ioctl_req, ioctl_start_dial_tone, ioctl_stop_tone, ioctl_start_ring,
	ioctl_stop_ring, ioctl_ring_back, ioctl_busy_tone, ioctl_exception, 
	ioctl_hook_state, ioctl_dtmf, ioctl_unsupported;
} MP_STAT;
#endif


/***********************
* PHONE DRIVER'S STUFF *
***********************/
static char MP_VER[] = "Version: 1.1";
static char MP_NAME[] = "Marvell Telephony Driver";

typedef struct {
	struct phone_device p;
	unsigned int ch_info;
	unsigned char ch;
	unsigned long reg_base;
	unsigned int irq;
	int readers, writers;
        wait_queue_head_t poll_q;
#ifdef MV_STAT
	/* statistics counters */
	MP_STAT stat;
#endif
} MV_PHONE;


static MV_PHONE mv_phone[TDM_CH_MAX];
#define get_mp(ch) (&mv_phone[ch])

/* Forward declarations */
static int __init mp_init(void);
static void __exit mp_exit(void);
static int mp_open(struct phone_device *pdev, struct file *file_p);
static int mp_close(struct inode *inode, struct file *file_p);
static ssize_t mp_read(struct file * file_p, char __user *buf, size_t length, loff_t * ppos);
static ssize_t mp_write(struct file *file_p, const char __user *buf, size_t count, loff_t * ppos);
static unsigned int mp_poll(struct file *file_p, poll_table * wait);
static int mp_ioctl(struct inode *inode, struct file *file_p, unsigned int cmd, unsigned long arg);
static irqreturn_t mp_int_handler(int irq, void *dev_id, struct pt_regs *regs);
#ifdef MV_STAT
static void mp_stat_show(MV_PHONE *mp);
#endif
#ifdef MV_IOCTL
extern MV_STATUS mvTdmSpiRead(MV_U32 tdmCh, MV_U8 addr, MV_U8 *data);
extern MV_STATUS mvTdmSpiWrite(MV_U32 tdmCh, MV_U8 addr, MV_U8 data);
extern MV_STATUS mvTdmChSpiTest(MV_U32 loop);
#endif

/* Module stuff */
module_init(mp_init);
module_exit(mp_exit);
MODULE_DESCRIPTION("Marvell Telephony Device Driver - www.marvell.com");
MODULE_AUTHOR("Tzachi Perelstein <tzachi@marvell.com>");
MODULE_LICENSE("GPL");

/* Driver's operations */
static struct file_operations mp_fops =
{
        .owner          = THIS_MODULE,
        .read           = mp_read,
        .write          = mp_write,
        .poll           = mp_poll,
        .ioctl          = mp_ioctl,
        .release        = mp_close,
};





static int __init mp_init(void)
{
	MV_PHONE *mp;
	int i;

	printk("%s, %s\n", MP_NAME, MP_VER);

	/* General TDM and SLIC init */
	mvTdmInit();
	mvTdmShowProperties();

	/* per channel init */
	for(i=0; i<TDM_CH_MAX; i++) {

		mp = get_mp(i);

		mp->p.board = mp->ch = i;
		mp->p.f_op = &mp_fops;
		mp->p.open = mp_open;

		mp->reg_base = FPGA_BASE;
		mp->irq = MP_IRQ;
        	init_waitqueue_head(&mp->poll_q);

		mvTdmChInit(&mp->p, i, &mp->ch_info);
		mvTdmChShowProperties(mp->ch_info);

		phone_register_device(&mp->p, PHONE_UNIT_ANY);
	
		printk("phone%d registered\n",i);
	}

	return 0;
}

static void __exit mp_exit(void)
{
	MV_PHONE *mp;
	int i;

	for(i=0; i<TDM_CH_MAX; i++) {
		mp = get_mp(i);

		mvTdmChRemove(mp->ch_info);

		mp->reg_base = 0;
		mp->irq = 0;
		mp->p.f_op = NULL;
		mp->p.open = NULL;
		mp->p.board = 0;
		phone_unregister_device(&mp->p);
		printk("phone%d removed\n",i);
	}
}

static int mp_open(struct phone_device *p, struct file *file_p)
{
	MV_PHONE *mp = get_mp(p->board);
	file_p->private_data = mp;

	printk("Opening phone channel %d - \n",mp->ch);

	if (!mp->reg_base) {
		printk("device not found\n");
		return -ENODEV;
	}

        if (file_p->f_mode & FMODE_READ) {
		if(!mp->readers) {
	                mp->readers++;
        	} else {
			printk("device is busy (read)\n");
                	return -EBUSY;
		}
        }

	if (file_p->f_mode & FMODE_WRITE) {
		if(!mp->writers) {
			mp->writers++;
		} else {
			if (file_p->f_mode & FMODE_READ){
				mp->readers--;
			}
			printk("device is busy (write)\n");
			return -EBUSY;
		}
	}

#ifdef MV_STAT
	memset(&mp->stat, 0, sizeof(MP_STAT));
#endif

	if(request_irq(mp->irq, mp_int_handler,
		SA_INTERRUPT, "MP", mp ) ) {
		printk("Failed to connect IRQ %d\n", mp->irq);
		mp->irq = 0;
		return -EIO;
	}
	MPDBG(MP_DBG_ISR, ("using IRQ %d\n",mp->irq));

	mvTdmChStart(mp->ch_info);

	return 0;
}

static int mp_close(struct inode *inode, struct file *file_p)
{
	MV_PHONE *mp = file_p->private_data;

	printk("Closing Marvell phone device");

	mvTdmChStop(mp->ch_info);

	free_irq(MP_IRQ, mp);

	if (file_p->f_mode & FMODE_READ)
		mp->readers--;
	if (file_p->f_mode & FMODE_WRITE)
		mp->writers--;

	file_p->private_data = NULL;

	return 0;
}

static ssize_t mp_read(struct file * file_p, char __user *buf, size_t length, loff_t * ppos)
{
	MV_PHONE * mp = file_p->private_data;/*get_mp(NUM(file_p->f_dentry->d_inode));*/
	MV_STATUS status;
	unsigned long count = 0;
	unsigned int err = 0;

	MPDBG(MP_DBG_RX, ("->%s\n",__FUNCTION__));
	MV_STAT_INC(mp->stat.read_req);

	disable_irq(mp->irq);

	count = min(length, (size_t)MV_TDM_BUFF_SIZE);
	status = mvTdmChRx(mp->ch_info, buf, count);

	if(status == MV_NOT_READY) { 
		MPDBG(MP_DBG_RX, ("read buff empty, try again.\n"));
		MV_STAT_INC(mp->stat.read_buff_empty);
		err = -EAGAIN;
	}
	else if(status == MV_FAIL) {
		printk("%s: copy to user failed.\n",__FUNCTION__);
		MV_STAT_INC(mp->stat.read_copy_to_user);
		err = -EFAULT;
	}
	else {
		MPDBG(MP_DBG_RX, ("%s: copy to user ok.\n",__FUNCTION__));
		err = 0;
	}

	MPDBG(MP_DBG_RX, ("<-%s\n",__FUNCTION__));
	enable_irq(mp->irq);
	return (err)?err:count;
}

static ssize_t mp_write(struct file *file_p, const char __user *buf, size_t length, loff_t * ppos)
{
	MV_PHONE *mp = file_p->private_data;
	MV_STATUS status;
	unsigned long count = 0;
	int err = 0;

	MPDBG(MP_DBG_TX, ("->%s\n",__FUNCTION__));
	MV_STAT_INC(mp->stat.write_req);

	disable_irq(mp->irq);
	
	count = min(length, (size_t)MV_TDM_BUFF_SIZE);
	status = mvTdmChTx(mp->ch_info, (MV_U8 *)buf, count);

	if(status == MV_NOT_READY) { 
		MPDBG(MP_DBG_TX, ("write buff full, try again.\n"));
		MV_STAT_INC(mp->stat.write_buff_full);
		err = -EAGAIN;
	}
	else if(status == MV_FAIL) {
		printk("%s: copy from user failed.\n",__FUNCTION__);
		MV_STAT_INC(mp->stat.write_copy_from_user);
		err = -EFAULT;
	}
	else {
		MPDBG(MP_DBG_TX, ("%s: copy from user ok.\n",__FUNCTION__));
		err = 0;
	}

	MPDBG(MP_DBG_TX, ("<-%s\n",__FUNCTION__));
	enable_irq(mp->irq);
	return (err)?err:count;
}

static unsigned int mp_poll(struct file *file_p, poll_table * wait)
{
	MV_PHONE *mp = file_p->private_data;/*get_mp(NUM(file_p->f_dentry->d_inode));*/
	unsigned int mask = 0;

	MPDBG(MP_DBG_POLL, ("-> %s\n",__FUNCTION__));
	MV_STAT_INC(mp->stat.poll_req);

	poll_wait(file_p, &(mp->poll_q), wait);

	disable_irq(mp->irq);

	if(mvTdmChExceptionReady(mp->ch_info)) {
		mask |= POLLPRI;		/* events */
		MPDBG(MP_DBG_POLL, ("poll can get event\n"));
		MV_STAT_INC(mp->stat.poll_can_do_event);
	}
	if(mvTdmChRxReady(mp->ch_info)) {
		mask |= POLLIN | POLLRDNORM;	/* readable */
		MPDBG(MP_DBG_POLL | MP_DBG_RX, ("poll can read\n"));
		MV_STAT_INC(mp->stat.poll_can_read);
	}
	if(mvTdmChTxReady(mp->ch_info)) {
		mask |= POLLOUT | POLLWRNORM;	/* writable */
		MPDBG(MP_DBG_POLL | MP_DBG_TX, ("poll can write\n"));
		MV_STAT_INC(mp->stat.poll_can_write);
	}

	MPDBG(MP_DBG_POLL, ("<-%s\n",__FUNCTION__));
	enable_irq(mp->irq);
	return mask;
}

static int mp_ioctl(struct inode *inode, struct file *file_p, unsigned int cmd, unsigned long arg)
{
	MV_PHONE *mp = get_mp(iminor(inode) & 0xf);
	int retval = 0;

	MPDBG(MP_DBG_IOCTL, ("->%s\n",__FUNCTION__));
	MV_STAT_INC(mp->stat.ioctl_req);

	/* check ioctls only root can use */
	if (!capable(CAP_SYS_ADMIN)) {
		return -EPERM;
	}

	disable_irq(mp->irq);

	switch (cmd) {
		case PHONE_REC_START:
		case PHONE_PLAY_START:
			MPDBG(MP_DBG_IOCTL, ("PHONE_REC/PLAY_START\n"));
			mvTdmChEnable(mp->ch_info);
			break;
		case PHONE_REC_STOP:
		case PHONE_PLAY_STOP:
			MPDBG(MP_DBG_IOCTL, ("PHONE_REC/PLAY_STOP\n"));
			mvTdmChDisable(mp->ch_info);
			break;
		case PHONE_DIALTONE:
			MPDBG(MP_DBG_IOCTL, ("PHONE_DIALTONE\n"));
			MV_STAT_INC(mp->stat.ioctl_start_dial_tone);
			mvTdmChDialTone(mp->ch_info);
			break;
		case PHONE_BUSY:
			MPDBG(MP_DBG_IOCTL, ("PHONE_BUSY\n"));
			MV_STAT_INC(mp->stat.ioctl_busy_tone);
			mvTdmChBusyTone(mp->ch_info);
			break;
		case PHONE_CPT_STOP:
			MPDBG(MP_DBG_IOCTL, ("PHONE_CPT_STOP\n"));
			MV_STAT_INC(mp->stat.ioctl_stop_tone);
			mvTdmChStopTone(mp->ch_info);
			break;
	 	case PHONE_RING_START:
			MPDBG(MP_DBG_IOCTL, ("PHONE_RING_START\n"));
			MV_STAT_INC(mp->stat.ioctl_start_ring);
			mvTdmChStartRing(mp->ch_info);
			break;
		case PHONE_RING_STOP:
			MPDBG(MP_DBG_IOCTL, ("PHONE_RING_STOP\n"));
			MV_STAT_INC(mp->stat.ioctl_stop_ring);
			mvTdmChStopRing(mp->ch_info);
			break;
		case PHONE_EXCEPTION:
		{
			MV_U8 offhook;
			union telephony_exception ex;
			MPDBG(MP_DBG_IOCTL, ("PHONE_EXCEPTION\n"));
			ex.bytes = 0;
			MV_STAT_INC(mp->stat.ioctl_exception);
			mvTdmChEventGet(mp->ch_info, &offhook);
			if(offhook)
				ex.bits.hookstate = 1;
			retval = ex.bytes;
		}
			break;
		case PHONE_RINGBACK:
			MPDBG(MP_DBG_IOCTL, ("PHONE_RINGBACK\n"));
			MV_STAT_INC(mp->stat.ioctl_ring_back);
			mvTdmChRingBackTone(mp->ch_info);
			break;
#if 0
		case PHONE_HOOKSTATE:
			MPDBG(MP_DBG_IOCTL, ("PHONE_HOOKSTATE\n"));
			MV_STAT_INC(mp->stat.ioctl_hook_state);
			retval = loopStatus();
			break;
		case PHONE_GET_DTMF:
			MPDBG(MP_DBG_IOCTL, ("PHONE_GET_DTMF\n"));
			MV_STAT_INC(mp->stat.ioctl_dtmf);
			retval = digit();
			break;
#endif

		case PHONE_MV_READ_REG:
			MPDBG(MP_DBG_IOCTL, ("PHONE_MV_READ_REG\n"));
#ifdef MV_IOCTL
			mvTdmSpiRead(mp->ch_info, arg, (MV_U8*)&retval);
#else
			printk("Driver is not configured to support this IOCTL\n");
			retval = -1;
#endif
			break;
		case PHONE_MV_WRITE_REG:
			MPDBG(MP_DBG_IOCTL, ("PHONE_MV_WRITE_REG\n"));
#ifdef MV_IOCTL
			mvTdmSpiWrite(mp->ch_info,(arg>>16)&0xff,arg&0xff);
#else
			printk("Driver is not configured to support this IOCTL\n");
			retval = -1;
#endif
			break;
		case PHONE_MV_SPI_TEST:
			MPDBG(MP_DBG_IOCTL, ("PHONE_MV_SPI_TEST\n"));
#ifdef MV_IOCTL
			mvTdmChSpiTest(10000);
#else
			printk("Driver is not configured to support this IOCTL\n");
			retval = -1;
#endif
			break;
		case PHONE_MV_STAT_SHOW:
			MPDBG(MP_DBG_IOCTL, ("PHONE_MV_STAT_SHOW\n"));
#ifdef MV_STAT
			mp_stat_show(mp);
#else
			printk("Driver is not configured to support statistics\n");
			retval = -1;
#endif
			break;
		default:
			MV_STAT_INC(mp->stat.ioctl_unsupported);
			printk("%s: Unsupported IOCTL\n",__FUNCTION__);
	}

	MPDBG(MP_DBG_IOCTL, ("<-%s\n",__FUNCTION__));
	enable_irq(mp->irq);
	return retval;
}


/* 
** Two types of interrupts exists: 
** (1) Slic async event (e.g. on/off-hook). 
** (2) FPGA sync 10ms event (data is ready for read+write)
*/
static irqreturn_t mp_int_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	MV_PHONE *mp = dev_id;
	MV_STAT_INC(mp->stat.int_total);

	MPDBG(MP_DBG_ISR, ("-> %s\n",__FUNCTION__));

	if(mvTdmIsr() == MV_OK) {
		MV_STAT_INC(mp->stat.int_wake_poll);
		wake_up_interruptible(&mp->poll_q);
	}
	else {
		MV_STAT_INC(mp->stat.int_empty);
		return IRQ_NONE;
	}

	return IRQ_HANDLED;
}

#ifdef MV_STAT
static void mp_stat_show(MV_PHONE *mp)
{
	MP_STAT *mps = &mp->stat;
	printk("\n\nMarvell Phone Device Driver Statistics (Informative/Warning/Error):\n");
	printk("===================================================================\n");
	printk("\nDriver version: %s\n",MP_VER);
	printk("\nDriver-Application interface:\n");
	printk("-----------------------------\n");
	printk("\n(I) Total read:  %d\n",mps->read_req);
	printk("(W) Read buffer is empty: %d\n",mps->read_buff_empty);
	printk("(E) Read blocking error: %d\n",mps->read_blocking);
	printk("(E) Read cpoy error: %d\n",mps->read_copy_to_user);
	printk("\n(I) Total write: %d\n",mps->write_req);
	printk("(W) Write buffer is full: %d\n",mps->write_buff_full);
	printk("(E) Write blocking error: %d\n",mps->write_blocking);
	printk("(E) Write cpoy error: %d\n",mps->write_copy_from_user);
	printk("\n(I) Total poll:  %d\n",mps->poll_req);
	printk("(I) Poll read:  %d\n",mps->poll_can_read);
	printk("(I) Poll write:  %d\n",mps->poll_can_write);
	printk("(I) Poll exception:  %d\n",mps->poll_can_do_event);
	printk("\n(I) Total ioctl: %d\n",mps->ioctl_req);
	printk("(E) Total unsupported ioctl: %d\n",mps->ioctl_unsupported);
	printk("\nDriver-HW interface:\n");
	printk("--------------------\n");
	printk("\n(I) Total interrupts: %d\n",mps->int_total);
	printk("(I) Async SLIC interrupts: %d\n",mps->int_slic);
	printk("(I) Sync Data interrupts: %d\n",mps->int_data);
	printk("\n(I) Write frames ok: %d\n",mps->int_write_ok);
	printk("(W) Write sync lost: %d\n",mps->int_write_sync_lost);
	printk("(W) Write buffer is empty: %d\n",mps->int_write_buff_empty);
	printk("\n(I) Read frames ok: %d\n",mps->int_read_ok);
	printk("(W) Read sync lost: %d\n",mps->int_read_sync_lost);
	printk("(W) Read buffer is full: %d\n",mps->int_read_buff_full);
	printk("\n(I) Waiking poll: %d\n",mps->int_wake_poll);
	printk("(W) Nothing was done in interrupt: %d\n",mps->int_empty);
}
#endif

